
//  Copyright 2021 Aaron Hance
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

{ 
	Drivers->Storage->AHCI->AHCI - AHCI Driver.
	
	@author(Aaron Hance <ah@aaronhance.me>)
}
unit AHCI;

interface

uses 
    util,
    PCI,
    drivertypes,
    drivermanagement,
    lmemorymanager,
    console,
    vmemorymanager,
    AHCITypes,
    lists;

var
    ahciControllers : PDList;
    page_base : puint32;

procedure init();
function load(ptr : void) : boolean;

implementation

procedure init();
var
    devID : TDeviceIdentifier;
begin
    console.writestringln('AHCI: Registering driver');
    devID.bus:= biPCI;
    devID.id0:= idANY;
    devID.id1:= $00000001;
    devID.id2:= $00000006;
    devID.id3:= $00000001;
    devID.id4:= idANY;
    devID.ex:= nil;
    drivermanagement.register_driver('ATA/PI AHCI Driver', @devID, @load);
    //TODO check IDE devices in ide for sata devices
end;

procedure stop_port(port : PHBA_Port);
begin
    port^.cmd_sts := port^.cmd_sts and not $1; ///maybe also bit 4
    while (port^.cmd_sts and $1) = 1 do begin
        //wait for the port to stop
    end;
end;

procedure start_port(port : PHBA_Port);
begin
    port^.cmd_sts := port^.cmd_sts or $1; ///maybe also bit 4
    while (port^.cmd_sts and $1) = 0 do begin
        //wait for the port to start
    end;
end;


{
    Check the ports on the controller and setup the command list, FIS, and command table
}
procedure check_ports(controller : PAHCI_Controller);
var
    i : uint32;
    port : PHBA_Port;
    device : PAHCI_Device;
    cmd_list_base : puint32;
    fis_base : puint32;
    cmd_table_base : puint32;
    cmd_header : PHBA_CMD_HEADER;
begin

    for i:=0 to 31 do begin
        if (controller^.mio^.port_implemented shr i) = 1 then begin
            port := @controller^.mio^.ports[i];
            console.writestring('AHCI: Port ');
            console.writeint(i);
            console.writestring(' implemented on controller ');
            console.writehexln(uint32(controller^.pci_device^.address5));

            //NEEED TO STOP the port before doing anything

            device := PAHCI_Device(DL_Add(controller^.devices));
            device^.port := port;

            //allocate memory for the command list and ensure it is aligned to 1024 bytes
            cmd_list_base := kalloc(sizeof(THBA_CMD_HEADER) * 64);
            cmd_list_base := puint32((uint32(cmd_list_base) + 1023) and $FFFFFC00);

            //allocate memory for the FIS and ensure it is aligned to 256 bytes
            fis_base := kalloc(sizeof(THBA_FIS) * 8);
            fis_base := puint32((uint32(fis_base) + 255) and $FFFFFF00);

            //set the command list base address
            port^.cmd_list_base := uint32(vtop(cmd_list_base)); //todo set virtual address in device
            port^.cmd_list_base_upper := 0;

            memset(uint32(cmd_list_base), 0, sizeof(THBA_CMD_HEADER));

            device^.command_list := PHBA_CMD_HEADER(cmd_list_base);

            //set the FIS base address
            port^.fis_base := uint32(vtop(fis_base));
            port^.fis_base_upper := 0;

            memset(uint32(fis_base), 0, sizeof(THBA_FIS));

            device^.fis := PHBA_FIS(fis_base);
            //todo check how many simultaneous commands are supported

            //allocate memory for the command table and ensure it is aligned to 1024 bytes
            cmd_table_base := kalloc(sizeof(THBA_CMD_TABLE) * 64);
            cmd_table_base := puint32((uint32(cmd_table_base) + 1023) and $FFFFFC00);

            device^.command_table := PHBA_CMD_TABLE(cmd_table_base);

            //set the command table base address and setup command table
            for i:=0 to 31 do begin
                //set command header locations
                command_header := PHBA_CMD_HEADER(uint32(cmd_list_base) + (i * sizeof(THBA_CMD_HEADER)));

                command_header^.prdtl := 32;
                //TODO do i need to set prdbc byte count here?
                command_header^.cmd_table_base := uint32(vtop(cmd_table_base) + (i * sizeof(THBA_CMD_TABLE));
                command_header^.cmd_table_base_upper := 0;
                memset(uint32(cmd_table_base) + (i * sizeof(THBA_CMD_TABLE)), 0, sizeof(THBA_CMD_TABLE));
            end;

            start_port(port);
        end;
    end;


end;

// procedure identify_device(controller : PAHCI_Controller; port : uint32);
// var
//     fis : PFIS_REG_H2D;
//     cmd : uint32;
//     i : uint32;
// begin
// end;

function load(ptr : void) : boolean;
var 
    device : PPCI_Device;
    controller : PAHCI_Controller;
    i : uint32;
    base : PHBA_Memory;
    cmd_list_base : puint32;
    fis_base : puint32;
    cmd_table_base : puint32;
begin
    console.writestringln('AHCI: initilizing a new controller');

    if ahciControllers = nil then begin
        console.writestringln('AHCI: Initializing controller list');
        ahciControllers := DL_New(SizeOf(TAHCI_Controller));
    end;

    device := PPCI_Device(ptr);

    controller := PAHCI_Controller(DL_Add(ahciControllers));
    controller^.devices := DL_New(SizeOf(TAHCI_Device));

    controller^.pci_device := PPCI_Device(device);


    //get the base address of the controller
    page_base := kpalloc(device^.address5); // TODO MEM memory manager need to be improved
    base := PHBA_Memory(device^.address5);
    controller^.mio := base;

    console.writestring('AHCI: Controller at: ');
    console.writehexln(device^.address5);
    console.writehexln(uint32(page_base));

    //enable AHCI mode, TODO check if is not already in AHCI mode and enable it, also perhaps remove if loaded as IDE
    base^.global_ctrl := base^.global_ctrl or AHCI_CONTROLLER_MODE;

    //clear any pending interrupts
    base^.int_status := $FFFFFFFF;


    

    check_ports(controller);




    // console.writestringln('AHCI: initilizing a new controller');
    // ahciControllers[ahciControllerCount] := ptr;
    // hba[ahciControllerCount] := PPCI_Device(ptr)^.address5;
    // new_page_at_address(hba[ahciControllerCount]);

    // //here would be any controller setup needed

    // check_ports(ahciControllerCount);
    // ahciControllerCount += 1;
    // exit(true);
end;



end.
