//  Copyright 2021 Aaron Hance
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

{ 
	Drivers->Storage->AHCI->AHCITypes - AHCI Driver Types.
	
	@author(Aaron Hance <ah@aaronhance.me>)
}
unit AHCITypes;
interface
uses 
    util,
    PCI,
    drivertypes,
    drivermanagement,
    lmemorymanager,
    console,
    vmemorymanager,
    idetypes,
    lists;

const
    AHCI_CONTROLLER_MODE = $80000000;

    CMD_LIST_ALIGN = 1024;
    CMD_TBL_ALIGN = 128;

    NUM_CMD_ENTRIES = 32;

    //device type signatures
    SATA_SIG_ATA = $00000101;
    SATA_SIG_ATAPI = $EB140101;
    SATA_SIG_SEMB = $C33C0101;
    SATA_SIG_PM = $96690101;



type

    { 
        AHCI Host Bus Adapter (HBA) Memory-Mapped Register Interface
        This structure is used to access the AHCI HBA's memory-mapped registers. 
        The AHCI HBA's memory-mapped registers are used to configure the HBA and
        to issue commands to the SATA devices connected to the HBA.

        The AHCI HBA's memory-mapped registers are accessed by reading and writing
        to the HBA's memory-mapped I/O space. The HBA's memory-mapped I/O space is
        typically mapped to a region of physical memory by the system's BIOS or
        UEFI firmware. The HBA's memory-mapped I/O space is typically mapped to a
        region of physical memory that is accessible to the system's CPU(s) and
        other devices.
    }
    THBA_Port = bitpacked record 
        cmdl_basel: uint32;     // Command List Base Address Lower 32-bits
        cmdl_baseu: uint32;     // Command List Base Address Upper 32-bits
        fis_basel: uint32;      // FIS Base Address Lower 32-bits
        fis_baseu: uint32;      // FIS Base Address Upper 32-bits
        int_status: uint32;     // Interrupt Status
        int_enable: uint32;     // Interrupt Enable
        cmd: uint32;            // Command and Status
        rsv0: uint32;           // Reserved
        tfd: uint32;            // Task File Data
        signature: uint32;      // Signature
        sata_stat: uint32;      // SATA Status (SCR0:SStatus)
        sata_ctrl: uint32;      // SATA Control (SCR2:SControl)
        sata_error: uint32;     // SATA Error (SCR1:SError)
        sata_active: uint32;    // SATA Active
        cmd_issue: uint32;      // Command Issue
        sata_noti: uint32;      // SATA Notification
        fis_switch_ctrl: uint32;// FIS-based Switch Control
        rsv1: array[0..10] of uint32;
    end;

    PHBA_Port = ^THBA_Port;

    { 
        AHCI Host Bus Adapter (HBA) Memory-Mapped Register Interface
        This structure is used to access the AHCI's memory-mapped registers.
        The AHCI's memory-mapped registers are used to configure the HBA and to
        issue commands to the SATA devices connected to the HBA.
    }
    THBA_Memory = bitpacked record 
        capabilites: uint32;    // Host Capabilities
        global_ctrl: uint32;    // Global Host Control
        int_status: uint32;     // Interrupt Status
        ports_implimented: uint32;// Ports Implemented 
        version: uint32;        // Version
        ccc_control: uint32;    // Command Completion Coalescing Control
        ccc_ports: uint32;      // Command Completion Coalescing Ports
        em_location: uint32;    // Enclosure Management Location
        em_control: uint32;     // Enclosure Management Control
        capabilites2: uint32;   // Host Capabilities Extended
        bohc: uint32;           // BIOS/OS Handoff Control and Status
        rsv0: array[0..$210] of boolean;
        ports: array[0..31] of THBA_Port;
    end;

    PHBA_Memory = ^THBA_Memory;

    { 
        AHCI Host Bus Adapter (HBA) FIS (Frame Information Structure) Interface
        This structure is used to access the AHCI HBA's FIS (Frame Information Structure)
        memory-mapped registers. 
    }
    THBA_FIS = bitpacked record 
        dsfis: array[0..$1F] of uint32; // DMA Setup FIS
        rsv0: array[0..$1F] of uint32;
        psfis: array[0..$1F] of uint32; // PIO Setup FIS
        rsv1: array[0..$1F] of uint32;
        rfis: array[0..$1F] of uint32;  // D2H Register FIS
        rsv2: array[0..$1F] of uint32;
        sdbfis: array[0..$F] of uint32; // Set Device Bits FIS
        ufis: array[0..$1F] of uint32;  // Unknown FIS
        rsv3: array[0..$67] of uint32;
    end;

    PHBA_FIS = ^THBA_FIS;

    { 
        AHCI Host Bus Adapter (HBA) Command Header Interface
        This structure is used to access the AHCI HBA's Command Header memory-mapped
        registers.
    }
    THBA_CMD_HEADER = bitpacked record 
        cmd_fis_length: uint8;   // Command FIS Length
        atapi: uint8;            // ATAPI
        write: uint8;            // Write
        prefetchable: uint8;     // Prefetchable
        reset: uint8;            // Reset
        bist: uint8;             // BIST
        clear_busy: uint8;       // Clear Busy
        reserved0: uint8;        // Reserved
        prdtl: uint16;           // Physical Region Descriptor Table Length
        prdbc: uint32;           // Physical Region Descriptor Byte Count
        cmd_table_base: uint32;  // Command Table Base Address
        cmd_table_baseu: uint32; // Command Table Base Address Upper 32-bits
        rsv0: array[0..4] of uint32;
    end;

    PHBA_CMD_HEADER = ^THBA_CMD_HEADER;

    { 
        AHCI Host Bus Adapter (HBA) Command Table Interface
        This structure is used to access the AHCI HBA's Command Table memory-mapped
        registers. The AHCI HBA's Command Table memory-mapped registers are used to
        configure the HBA and to issue commands to the SATA devices connected to the HBA.
    }
    THBA_CMD_TABLE = bitpacked record 
        cmd_fis: THBA_FIS;       // Command FIS
        acmd: array[0..$1F] of uint8; // ATAPI Command
        rsv0: array[0..$30] of uint8;
    end;

    PHBA_CMD_TABLE = ^THBA_CMD_TABLE;

    TCMD_LIST = array[0..255] of THBA_CMD_HEADER;
    PCMD_LIST = ^TCMD_LIST;

    { 
        AHCI Host Bus Adapter (HBA) Command Table Interface
        This structure is used to access the AHCI HBA's Command Table memory-mapped
        registers. 
    }
    THBA_PRD = bitpacked record 
        dba: uint32;             // Data Base Address
        dbau: uint32;            // Data Base Address Upper 32-bits
        rsv0: uint32;            // Reserved
        dbc: uint32;             // Data Byte Count
        rsv1: uint32;            // Reserved
    end;

    { 
        AHCI Host Bus Adapter (HBA) Command Table Interface
        This structure is used to access the AHCI HBA's Command Table memory-mapped
        registers. 
    }
    THBA_CMD = bitpacked record 
        header: THBA_CMD_HEADER; // Command Header
        table: THBA_CMD_TABLE;   // Command Table
        prd: array[0..31] of THBA_PRD; // Physical Region Descriptor Table
    end;

    { 
        AHCI Host Bus Adapter (HBA) Command Table Interface
        This structure is used to access the AHCI HBA's Command Table memory-mapped
        registers. 
    }
    THBA = bitpacked record 
        memory: THBA_Memory; // HBA Memory
        cmd: array[0..$7FF] of THBA_CMD; // Command List
    end;

    PHBA = ^THBA;

    //////////////////////////////////////////
    //////////// Asuro AHCI types ////////////
    //////////////////////////////////////////

    {
        Device type enum
    }
    TDeviceType = (
        SATA,
        ATAPI,
        SEMB,
        PM
    );

    {
        AHCI device reference
    }
    TAHCI_Device = bitpacked record
        port : PHBA_Port;
        device_type : TDeviceType;
        ata_info : TIdentResponse;
        command_list : PHBA_CMD_HEADER;
        fis : PHBA_FIS;
        command_table : PHBA_CMD_TABLE;
    end;

    PAHCI_Device = ^TAHCI_Device;

    {
        controller reference
    }
    TAHCI_Controller = bitpacked record
        pci_device : PPCI_Device;
        mio: PHBA_Memory;
        ata_info : TIdentResponse;
        devices : PDList;
    end;

    PAHCI_Controller = ^TAHCI_Controller;

     
 function get_device_type(sig : uint32) : TDeviceType;
//  function get_device_type_string(deive_type : TDeviceType) : string;

implementation

function get_device_type(sig : uint32) : TDeviceType;
begin
    case sig of
        SATA_SIG_ATA: begin
            get_device_type := SATA;
        end;
        SATA_SIG_ATAPI: begin
            get_device_type := ATAPI;
        end;
        SATA_SIG_SEMB: begin
            get_device_type := SEMB;
        end;
        SATA_SIG_PM: begin
            get_device_type := PM;
        end;
    end;
end;

// function get_device_type_string(deive_type : TDeviceType) : string;
// begin
//     case deive_type of
//         SATA: begin
//             get_device_type_string := 'SATA';
//         end;
//         ATAPI: begin
//             get_device_type_string := 'ATAPI';
//         end;
//         SEMB: begin
//             get_device_type_string := 'SEMB';
//         end;
//         PM: begin
//             get_device_type_string := 'PM';
//         end;
//     end;
// end;

end.

